# Python 2 and 3:
try:
    # Python 3 only:
    from urllib.parse import urlencode, urlsplit, parse_qs, unquote
except ImportError:
    # Python 2 only:
    from urlparse import parse_qs, urlsplit
    from urllib import urlencode, unquote



from datetime import datetime
from operator import itemgetter
from collections import OrderedDict

from django.core.urlresolvers import resolve,reverse
from django.http.request import QueryDict
from django.utils.http import is_safe_url
from django.conf import settings
from django.http.response import *

from rest_framework.views import APIView as GAPIView
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.request import Request
from rest_framework import generics

from .mixins import RelatedView
from .utility import is_ajax
from .py2_3 import *

class ListAPIView(generics.ListAPIView,RelatedView):

    def list(self,request,*args,**kwargs):
        """ 
        Overridden generics.ListAPIView list method to provide additional 
        functionality of related views data fetching and applied filters addition
        """
        response=super(ListAPIView,self).list(request,*args,**kwargs)
        #add applied_filters to the response which is set when filter_queryset method is called
        response=self.addAppliedFilters(response)
        #fetch data from the related views
        return self.fetch_related(request,response,*args,**kwargs)

    def addAppliedFilters(self,response):
        """
        Add the filters applied to the view to response using the view applied_filters attribute accessible with filters key

        """
        if hasattr(self,'applied_filters') and self.applied_filters:
            if not isinstance(response.data,(list,tuple)):
                response.data['filters']=self.applied_filters
        return response

    def filter_queryset(self,queryset):
        """ 
        Overridden generics.ListAPIView filter_queryset method for adding the filters applied to this view.
        Appends filters applied to ListAPIView instance as applied_filters attribute.
        It fetches the filter from filter_backends by calling its get_applied_filters method.
        """

        filters = {}
        for backend in list(self.filter_backends):
            backendobj = backend()
            queryset = backendobj.filter_queryset(self.request, queryset, self)
            if hasattr(backendobj,'get_applied_filters'):
                filters.update(backendobj.get_applied_filters())
        self.applied_filters = OrderedDict()
        from datetime import datetime, date
        for key,value in filters.items():
            if isinstance(value,(datetime,date)):
                self.applied_filters[key]=value
                del filters[key]
        self.applied_filters.update(sorted(filters.items(),key=itemgetter(1),reverse=True))
        return queryset


class RetrieveAPIView(generics.RetrieveAPIView,RelatedView):

    def retrieve(self,request,*args,**kwargs):
        """ 
        Overridden generics.RetrieveAPIView retrieve method to provide additional 
        functionality of related views data fetching
        """
        response=super(RetrieveAPIView,self).retrieve(request,*args,**kwargs)
        return self.fetch_related(request,response,*args,**kwargs)

class APIView(GAPIView,RelatedView):
    """
        Basic View which inherited from APIView which supports only related data fetching and response format support.

    """
    template_name = None
    def get(self,request,*args,**kwargs):
        response = Response({})
        return self.fetch_related(request,response,*args,**kwargs)

class JSONAPIView(APIView):
    renderer_classes = (JSONRenderer,)

class AttachedTabAPIView(RelatedView):
    """
    View attached to a TabAPIView,sharing the same base template.
    """
    def fetch_related(self,request,response,*args,**kwargs):
        response = super(AttachedTabAPIView,self).fetch_related(request,response,*args,**kwargs)
        if getattr(response,'data',None) and getattr(self,'_currenttab',None):
            response.data['current_tab'] = self._currenttab
        return response
    
    
class TabAPIView(RelatedView):
    defaulttab = None
    def fetch_related(self,request,response,*args,**kwargs):
        response = super(TabAPIView,self).fetch_related(request,response,*args,**kwargs)
        if getattr(response,'data',None) and getattr(self,'_currenttab',None):
            response.data['current_tab'] = self._currenttab
        return response

    def get_requested_views(self,request,returnformat):
        """get requested views from request.query_params.relview"""
        if isinstance(request,Request):
            relview_key = getattr(self,'relview_key','relview')
            reqviews = request.query_params.get(relview_key)
            if reqviews is None:
                # if no relview is passed in query_params then fetch
                # from its attribute relview which is set in urls.py
                if self.relview is not None:
                    reqviews = self.relview
            # if related views are not fetched either through query_params
            # or in urls.py and is not jsonrequest then fetch all views
            if reqviews is None: #and returnformat != 'json':
                tabmap = getattr(self,'tabmap',None)
                tabkey = getattr(self,'tab_key','tab')
                if tabmap:
                    currenttab = request.query_params.get(tabkey)
                    if not currenttab:
                        if self.defaulttab is not None:
                            currenttab = self.defaulttab

                    requestedtab = tabmap.get(currenttab,self.defaulttab)
                    if currenttab and requestedtab:
                        reqviews = map(lambda x: x.strip(),requestedtab.strip(',').split(','))
                        self._currenttab = currenttab
            if reqviews:
                return reqviews
        return super(TabAPIView,self).get_requested_views(request,returnformat)
                    
class FormView(APIView,RelatedView):
    def get(self,request,*args,**kwargs):
        self.processRequest(request,*args,**kwargs)
        self.is_get_call=True
        if self.get_form_data():
            last_flow=self._get_last_flow()
            if last_flow['by']==self.view_url_name:
                self._get_last_flow(pop=True)
            return self.postForm(request,*args,**kwargs)
        return self.getForm(request,*args,**kwargs)

    #@method_decorator(csrf_exempt)
    def post(self,request,*args,**kwargs):
        self.is_get_call=False
        self.processRequest(request,*args,**kwargs)
        return self.postForm(request,*args,**kwargs)
    def _get_urlname(self,url):
        urlname= None
        try:
            path=urlparse(url).path
            urlname = resolve(path).url_name
        except:
            pass
        return urlname
    def _get_last_flow(self,pop=False):
        last_flow= None
        flow = self.get_flow()
        if flow:
            if len(flow)>0:
                last_flow =flow[-1] if not pop else flow.pop()
            if len(flow)==0:
                self._destroy_flow(self.request)
        return last_flow

    def _get_current_referer(self,request):
        referer = request.META.get('HTTP_REFERER',None)
        next_url = request.GET.get('next')
        url_name = None
        #if not referer it means its a direct request
        if referer and is_safe_url(url = referer, host=request.get_host()):
            url_name = self._get_urlname(referer)
        if not url_name:
            referer=None
        if next_url:
            url_name,referer = 'homepage',next_url
        return (url_name,referer)


    def redirect_to(self,url,data={}):
        if is_ajax(self.request):
            resdata = {'location':url}
            resdata.update(data)
            return Response(resdata)
        return Response({},status=302,headers={'Location':url})

    def send_next(self,name,packet=None,query_dict={},returnurl=True,finalurl=None,data={}):
        """
        sends to next page maintaing history in session.
        - name : Name of the form view to be sent next without _form_view
        - packet : packet to be transferred to next form view
        - query_dict : query parameter to be passed
        - returnurl :
            True - return back to this view once the job has been done
            False - No need to return here just continue with normal flow
            <urlstr> - send to urlstr once the job finished. It must be a form_view url
        - finalurl - the final url to send.it resets the whole flow

        """
        view_url = name + '_form_view'
        try:
            urlpath = reverse(view_url)
            query_dict.update({'_caller':self.view_url_name})
            url = urlpath + '?'+ urlencode(query_dict)
        except:
            #@TODO need to clean every session data here
            url = reverse('homepage')
        if packet and isinstance(packet,dict):
            packet['to']=view_url
            self.request.session['_packet']=packet

        if finalurl:
            self._initiate_flow(self.request,finalurl) 
        elif returnurl == True:
            self._addurl(self._get_current_url())
        elif returnurl != False:
            self._addurl(returnurl)

        return self.redirect_to(url,data)

    def send_back(self,packet=None,data={},query_dict={}):
        obj = self._get_last_flow(pop=True)
        if obj:
            is_last = not self.get_flow()
            redirect_url = obj['url']
            if not is_last:
                query_dict.update({'_caller':self.view_url_name})
                if packet and isinstance(packet,dict):
                    packet['to']=self._get_urlname(redirect_url)
                    self.request.session['_packet']=packet
            redirect_url = self.get_query_url(redirect_url,query_dict)
            return self.redirect_to(redirect_url,data)
        return self.redirect_to(reverse('homepage'),data)

    def get_query_url(self,current_path,query_params={}):
        join_char = '&' if urlparse(current_path).query else '?'
        query_url = current_path + join_char + urlencode(query_params)
        return query_url
 
    def get_view_cache(self,key=None,pop=False):
        if not key:
            return self._sessdata
            
        data = self._sessdata.get(key,None)
        if pop:
            self.set_view_cache(key,None)
        return data


    def set_view_cache(self,key,data=None):
        if not key:
            raise Exception('Key is mandatory for setting form view cache')
        if data is None:
            self._sessdata.pop(key,None)
        else:
            self._sessdata[key]=data
        self.request.session[self.view_url_name]=self._sessdata
        return True

    def set_form_data(self,data):
        return self.set_view_cache('_formdata',data)

    def get_form_data(self):
        if not hasattr(self,'_formdata'):
            self._formdata=self.get_view_cache('_formdata')
        return self._formdata

    def get_caller(self):
        return self._caller

    def get_packet(self):
        return self._packet
    def get_flow(self):
        return self.request.session.get('_formflow',None)

    def _addurl(self,url):
        formflow = self.request.session.get('_formflow',[])
        #@TODO Remove it from here
        if not formflow:
            self._initiate_flow(self.request,withurl='/')
            formflow = self.request.session.get('_formflow',[])
            #raise Exception('_addurl called before initiating the flow')
        formflow.append({'by':self.view_url_name,'url':url})
        self.request.session['_formflow']=formflow
        return True

    def _destroy_flow(self,request):
        sess=request.session
        keys = request.session.keys()
        #@TODO check if these deletions are making multiple db hits
        if '_packet' in keys:
            del sess['_packet']
        if '_formflow' in keys:
            del sess['_formflow']
        for key in keys:
            if key.endswith('_form_view'):
                del sess[key]


    def _initiate_flow(self,request,withurl=None):
        self._destroy_flow(request)
        sess=request.session
        if not withurl:
            withurl = settings.LOGIN_REDIRECT_URL
        sess['_formflow']=[{'by':'initiator','url':withurl}]
        return True
    def _get_current_url(self):
        return reverse(self.view_url_name)

    def processRequest(self,request,*args,**kwargs):
        keyname = self.view_url_name
        self._caller = None
        self._packet = {}
        caller = request.query_params.get('_caller',False)
        #check if called from another view
        if caller:
            packet = request.session.get('_packet',{})
            #@TODO what to do if packet is not found
            # Is this invalid packet?
            if packet.get('to',None) == keyname:
                self._packet = packet
                request.session.pop('_packet',None)
            self._caller = caller
        else:
            url_name,referer = self._get_current_referer(request)
            if not referer:
                #@TODO A case when user refreshes the page the referer is not set in which get this gets called.Need to fix it later
                self._initiate_flow(request)
            else:
                if url_name == self.view_url_name:
                    #it is a request from its post form
                    pass
                elif url_name.endswith('_form_view'):
                    #the flow is broken and the user has switched to other flow
                    pass
                else:
                    self._initiate_flow(request,withurl=referer)
                    #it means urls in session are valid
        self._sessdata = request.session.get(keyname,{})

    def call_get(self,request,*args,**kwargs):
        if is_ajax(request):
            return Response(kwargs)
        return self.getForm(request,*args,**kwargs)




from django.conf import settings
from django.http.response import *

from rest_framework.response import Response

from .utility import subtractlists,Memoized,DummyRequest
from .py2_3 import *

AS_MAIN=1

class RelatedView(object):
    '''View class makes a view callable from other view. '''
    relview = None
    jointrel = None
    @classmethod
    def as_data(cls,**initkwargs):
        def view(request,*args,**kwargs):
            self=cls(**initkwargs)
            self.retType = 'data'
            self.request = request
            self.format_kwarg = None
            self.args = args
            self.kwargs = kwargs
            resp =  self.get(request,*args,**kwargs)
            if isinstance(resp,Response):resp = resp.data
            return resp

        setattr(view,'__name__',cls.__name__)
        setattr(view,'_class',cls)
        setattr(view,'_initkwargs',initkwargs)
        disablecache = getattr(settings,'MEMO_CACHE_DISABLED',False)
        if not disablecache and initkwargs.get('memoization',getattr(cls,'memoization',False)):
            view = Memoized(view)
        return view


    #def set_related_params(self,params,request,responsedata):
    def set_related_params(self,request,responsedata):
        """ 
        set parameter of related views. Will be available in
        related views kwargs as well as in request.query_params
        of request object passed to the related view
        """
        pass

    def set_pipelined_response(self,view_name,request,responsedata):
        """
        to communicate between different related views use this.
        Responsedata gets cascaded with responses of related views         called in related_views in views.py.
        """
        pass

    def get_final_response(self,request,response):
        """
        it gets the final response after related views have been executed and final response has been appended.
        """
        return response

    def updatekwargs(self,request):
        """
        to update kwargs with query params with greater priority of kwargs.
        """
        updated_dict={}
        requrl = request.build_absolute_uri()
        urlparts = urlsplit(requrl)
        unquote_query = unquote(urlparts.query)
        updated_dict = dict(parse_qs(unquote_query))
        updated_dict = { k:','.join(v) for k,v in updated_dict.items()}
        updated_dict.update(self.kwargs)

        self.kwargs = updated_dict


    def fetch_related(self,request,response,*args,**kwargs):
        ''' fetches data of related views and append to the result '''
        relateddata = {}

        if not isinstance(response.data,dict):
            response = self.get_final_response(request,response)
            return response
        # Retrieve the format of the response
        retformat=request.accepted_renderer.format


        #if hasattr(self,'retType') and self.retType=='data':
            #return response.data
        if not hasattr(self,'related_views') or len(self.related_views)==0:
            response = self.get_final_response(request,response)
            return response

        reqviews = self.get_requested_views(request,retformat)

        if reqviews:
            #check if there is any related view
            if isinstance(self.related_views,dict):
                """
                    A dummy request object is created here which has necessary attributes
                    of original request.The idea is if we pass the original request,then its query_params
                    attribute may have some keys which is common in related views but we dont want to
                    pass it to related view( most common page). The arguments to the related view is 
                    controlled by explicitly passing the name of view separated by colon(:) and the parameter
                    to be passed. e.g self.filter_params['topics:page']=request.query_params.get('page',None)
                """
                dummyreq = DummyRequest(request)
                self.updatekwargs(request)
                self.set_related_params(request,response.data)
                #self.set_related_params(updated_dict,request,response.data)
                for name in reqviews:
                    self.set_pipelined_response(name,request,response.data)
                    relobj = self.related_views.get(name,None)
                    dummyreq.query_params={}
                    if relobj:
                        #check that a name and handler function has been provided
                        if len(relobj)<1:
                            raise Exception('Related View must have a handler function')

                        callback = relobj[0]
                        #set the filters for this view which is passed in query_params attribute of request i.e dummyreq
                        if isinstance(relobj[1],str):
                            dummyreq.query_params = self.get_related_params(relobj[1],name)
                        resp = callback(dummyreq,**dummyreq.query_params)
                        if resp is None:
                            raise Exception('The response must be of type Response,dict,list. None received')
                        if type(resp) == Response:
                            resp = resp.data
                        if len(relobj)>2 and relobj[2]==1:
                            response.data[name]=resp
                        elif len(relobj)>1 and not isinstance(relobj[1],str) and relobj[1]==1:
                            response.data[name]=resp
                        else:
                            relateddata[name]=resp
                response.data['extdata']=relateddata
        response = self.get_final_response(request,response)
        if not isinstance(response,(Response,HttpResponse)):
            raise Exception("Expected a django `Response` type to be returned from %s" %self.get_final_response.__name__)
        if hasattr(self,'retType') and self.retType=='data':
            return response.data
        return response

    def get_related_params(self,param_str,viewname):
        related_params = {}
        param_str = param_str.strip(',')
        parameters  = param_str.split(',')
        for p in parameters:
            p = p.strip()
            if p == '*':
                related_params.update(self.kwargs)
            elif '=' in p:
                p_split = p.split('=')
                related_params[p_split[0]]=p_split[1]
            elif ' as ' in p:
                p_split = p.split(' as ')
                if self.kwargs.get(p_split[0]):
                    related_params[p_split[1]]=self.kwargs.get(p_split[0])
            else:
                if self.kwargs.get(p):
                    related_params[p]=self.kwargs.get(p)
        return related_params

    def get_requested_views(self,request,returnformat):
        """get requested views from request.query_params.relview"""
        relview_key = getattr(self,'relview_key','relview')
        reqviews = request.query_params.get(relview_key)
        if reqviews is None:
            # if no relview is passed in query_params then fetch
            # from its attribute relview which is set in urls.py
            if self.relview is not None:
                reqviews = self.relview
        # if related views are not fetched either through query_params
        # or in urls.py and is not jsonrequest then fetch all views
        if reqviews is None: #and returnformat != 'json':
            reqviews = 'all'
        if self.jointrel is not None:
            reqviews = reqviews+','+self.jointrel
        if reqviews:
            reqviews = reqviews.split(',')
            relviews = self.related_views.keys()
            include = []
            exclude = []
            for reqview in reqviews:
                if reqview == 'all':
                    include += relviews
                elif reqview[0] == '-':
                    exclude.append(reqview[1:])
                elif reqview not in include:
                    include.append(reqview)
            reqviews = subtractlists(include,exclude)
        return reqviews




import types, sys, copy, hashlib

from itertools import chain
from operator import itemgetter

from django.db import models
from django.http import Http404
from django.conf import settings
from django.core.cache import cache
from django.shortcuts import render
from django.core.urlresolvers import reverse
from django.core.exceptions import FieldError

from rest_framework import status
from rest_framework.response import Response 
from rest_framework.generics import GenericAPIView
from rest_framework.pagination import PageNumberPagination

from .py2_3 import *

def register_as_module(cls, mod_name):
    '''
    Create a dynamic module from public members of a given class
    and set it to sys modules
    '''
    _cls_module = types.ModuleType(mod_name)
    _defined_field = [attr for attr in dir(cls) if not attr.startswith("_")]
    for field in _defined_field:
        setattr(_cls_module,field, getattr(cls,field))
    sys.modules.setdefault('%s.%s'%(cls.__module__,mod_name),_cls_module)

def register_as_proxy_model(model, manager=models.Manager, module='common.models', app_label='cms', write=False):
    '''
    Return a ProxyModel of a given Model if it is not already proxy and attach new manager object
    otherwise return same  model and also keep default manager as '_objects' attribute
    '''
    if model._meta.proxy:
        return model
    manager_obj = manager() if isinstance(manager,models.Manager) else manager(model=model)
    app_label = model._meta.app_label if write else app_label
    _Meta = type('Meta',(),{'proxy':True,'app_label':app_label})
    attrs = {
            'Meta': _Meta,
            '__module__': module,
            'objects': manager_obj,
            '_default_manager': manager_obj,
    }
    Klass = type('Proxy%s'%model.__name__,(model,),attrs)
    Klass._objects = model.objects
    return Klass

class BasicModelManager(models.Manager):
    '''
    Model Manager designed for basic business filters
    applied in all models of the project
    '''
    def __init__(self,*args,**kwargs):
        self._model = kwargs.pop('model',self.__class__)
        self.fields = self._model._meta.get_all_field_names()

        self.filters = dict()
        self._set_filter(show_on_site = True)
        self._set_filter(site = settings.SITE_ID)
        self._set_filter(sites = settings.SITE_ID)

        return super(BasicModelManager, self).__init__(*args,**kwargs)

    def _set_filter(self,*args,**kwargs):
        field = kwargs.keys().pop()
        if field in self.fields:
            self.filters.update(kwargs)

    def get_queryset(self):
        queryset = super(BasicModelManager, self).get_queryset()
        try:
            return queryset.filter(**self.filters)
        except FieldError:
            return queryset

    def __repr__(self):
        return '<%s Manager object from %s.%s>' %(self._model.__name__,self.__module__,self.__class__.__name__)

def cstolist(value):
    """convert comma separated values to list identifying proper type.List object remains unchanged """
    if value in (None,'',u''):
        return []
    if isinstance(value,(str,unicode)):
        #do typecasting
        rawvalues=value.split(',')
        if rawvalues[0].isdigit():
            value = [int(v) for v in rawvalues]
        else:
            value = [v for v in rawvalues]
    elif isinstance(value,(int,)):
        value=[value]
    return value

def choicestodict(data,keys=None):
    '''
    convert django choices to list of dict by default containing text and value key.Pass keys for customization.
    Usage:
        a=[[1,'a'],[2,'b']]
        x=choicestodict(a)
        x = [{'value':1,'text':'a'},{'value':2,'text':'b'}]

        a=[[1,'a'],[2,'b']]
        x=choicestodict(a,['id','name'])
        x = [{'id':1,'name':'a'},{'id':2,'name':'b'}]

    Possible Use Cases:
        convert model choices to list of text,value pairs
        convert lms api data to list of id,name value pairs
    '''
    if not keys:
        keys = ['value','text']
    return list(map((lambda x: dict(zip(keys,x))),data))

def subtractlists(include,exclude):
    return filter(lambda x: x not in exclude, include) 

def is_ajax(request):
    return request.query_params.get('ajax',None) or request.is_ajax()

class NoPagination(object):
    """ 
    Wraps data wrapped in a dict with key name 'results'.Return Response object.
    Use with pagination_class=NoPagination as view attribute 
    """
    display_page_controls = False
    def paginate_queryset(self,queryset,request,view=None):
        if isinstance(queryset,models.Manager):
            queryset=queryset.all()
        return list(queryset)

    def get_paginated_response(self,data):
        return Response({'results':data})

    def get_results(self, data):
        return data['results']

class Memoized(object):
    excluded_renderer = ('api','json')

    def __init__(self, func):
        self._func = func

    def __repr__(self):
        return self._func.__repr__()

    def _create_cache_key(self,args,kwargs,initkwargs):
        kwargs.update(initkwargs)
        kwargs.pop('format','')
        filters = urlencode(kwargs)
        cache_key = hashlib.sha1('%s:%s' %(self._func.__name__,filters)).hexdigest()
        return cache_key[:250]
    
    def _get_set_cache(self,args,kwargs,cache_key):
        value = self._func(*args,**kwargs)
        cache_duration = getattr(self._func._class,'cache_duration',settings.MEMOIZE_DURATION)
        cache.set(cache_key,value,cache_duration)
        return value

    def _memoize_renderer(self,request):
        return request.accepted_renderer.format not in self.excluded_renderer

    def __call__(self, *args, **kwargs):
        cache_key = self._create_cache_key(args,kwargs,self._func._initkwargs)
        cache_data = self._memoize_renderer(args[0]) and cache.get(cache_key)
        return cache_data or self._get_set_cache(args,kwargs,cache_key)

class DummyRequest(object):
    """  
        A substitute for rest_framework request object which provides mutable query_params attribute. 
        All the related views are passed DummyRequest instance.The request data that is passed to Related Views are inserted into ites query_params attribute.
    """
    def __init__(self,request,data={}):
        self._request = request
        self.query_params = {}
        self.isDummy = True
        self.data = data
    
    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            return getattr(self._request,name)




from setuptools import setup

setup(
  name = 'drf-related-views',
  packages = ['rest_framework_related'],
  version = '0.0.4',
  description = 'Related Views for Django Rest Framework',
  author = 'Fasih Ahmad Fakhri',
  author_email = 'fasihahmadfakhri@gmail.com',
  url = 'https://github.com/fasihahmad/django-rest-framework-related-views',
  download_url='https://github.com/fasihahmad/django-rest-framework-related-views/archive/v0.0.4.tar.gz',
  keywords = 'django rest framework related views',
  license='GPL-3.0',
  classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries'
    ],
  install_requires=[
        'Django>=1.6.3',
        'django-filter>=0.11.0',
        'djangorestframework>=3.2.4',
        'djangorestframework-filters>=0.10.2'
    ],
)



import copy
import django_filters

from datetime import datetime
from django.http.request import QueryDict
from django.core.exceptions import FieldError
from django.db.models.fields import FieldDoesNotExist

try:
    from django_filters.rest_framework import DjangoFilterBackend
except ImportError:
    # Import for supportable djangorestframework == 3.2.4
    from rest_framework.filters import DjangoFilterBackend

from .utility import cstolist, NoPagination

class ListFilter(django_filters.Filter):
    def filter(self,qs,value):
        if value not in (None,'',[]):
            #convart comma separated str to list
            value = cstolist(value)
            #check is more than one filters are to be applied.
            #This scenerio has been written considering only 'in' query
            #whose performance is quite low as compared to direct matching.
            # Use 'in' query only when multiple filters are to be applied on single fields otherwise direct matching'
            # For other usecases please consider this point in mind before extending
            if len(value)>1:
                return self.get_method(qs)(**{'%s__%s'%(self.name,self.lookup_expr):value})
            else:
                return self.get_method(qs)(**{'%s'%(self.name):value[0]})
        return qs

class ValueListFilter(django_filters.Filter):
    def filter(self,value_list,value):
        validated = all([isinstance(each,dict) for each in value_list])
        if not validated:
            raise Exception('Invalid ValuList. Value list means, list of dictionary, so simple')
        if value not in (None,'',[]):
            value = cstolist(value)
            value_list = filter((lambda x,y=value:x.get(self.name) in y),value_list)
        return value_list

class ValueList(list):
    def __init__(self,value_list):
        self.model = self.__class__
        self.value_list = value_list

    def all(self):
        return self.value_list

    class _meta(object):
        @staticmethod
        def get_field_by_name(*args,**kwargs):
            raise FieldDoesNotExist()
    

class ExcludeListFilter(django_filters.Filter):
    def filter(self,qs,value):
        #import pdb;pdb.set_trace()
        if value not in (None,'',[]):
            #convart comma separated str to list
            value = cstolist(value)
            #check is more than one filters are to be applied.
            #This scenerio has been written considering only 'in' query
            #whose performance is quite low as compared to direct matching.
            # Use 'in' query only when multiple filters are to be applied on single fields otherwise direct matching'
            # For other usecases please consider this point in mind before extending
            if len(value)>1:
                return qs.exclude(**{'%s__%s'%(self.name,self.lookup_type):value})
            else:
                return qs.exclude(**{'%s'%(self.name):value[0]})
        return qs

class CountBackend(object):
    """ Limits the count of result fetched.Use only at the end of filters.Resets the default pagination to NoPagination """
    def filter_queryset(self,request,queryset,view):
        if hasattr(view,'limit_key'):
            countkey = view.limit_key
        else:
            countkey = 'limit'
        limit = request.query_params.get(countkey,None)
        if limit is None or not str(limit).isdigit():
            if not hasattr(view,'limit') or view.limit==-1:
                return queryset
            else:
                limit=view.limit
        view.pagination_class=NoPagination
        if limit == 0:
            return []
        return queryset[:int(limit)]

class ExcludeBackend(object):
    def filter_queryset(self,request,queryset,view):
        exclude_key = getattr(view,'exclude_key','excludekey')
        exclude_val = getattr(view,'exclude_value','excludevalue')
        self.exclude_key = key = request.query_params.get(exclude_key,'id')
        self.exclude_val = values = request.query_params.get(exclude_val)
        if values:
            values = cstolist(values)
            exclude_kwargs = {'{0}__in'.format(key) : values}
            return queryset.exclude(**exclude_kwargs)
        return queryset
    def get_applied_filters(self):
        return {'excluded_{0}'.format(self.exclude_key):self.exclude_val}

class OrderBackend(object):
    def filter_queryset(self,request,queryset,view):
        order_key = getattr(view,'order_key','order')
        order_by_key = getattr(view,'order_by_key','order_by')

        self.order_by_clause = getattr(view,'order_by_clause',{})

        order_by_param = request.query_params.get(order_by_key)

        order_param = request.query_params.get(order_key,'default')
        order_by_tuple = self.get_ordering(order_param)

        if order_by_param:
            ordering = cstolist(order_by_param)
            queryset=queryset.order_by(*ordering)
            self._filters = {order_by_key:order_by_param}

        elif order_by_tuple:
            ordering = order_by_tuple
            queryset = queryset.order_by(*ordering)
            #@TODO Handle ordering for null fields in database
#            if self.field_sort and not "__" in self.field_sort:
#                field_sort = self.field_sort
#                max_value = getattr(queryset.last(),field_sort,None)
#                if max_value and not isinstance(max_value,datetime):
#                    queryset = queryset.extra(select={'ordering':"case when "+ field_sort +" is null then "+str(max_value)+" else " + field_sort + " end"})
#                    queryset = queryset.extra(order_by=['ordering'])
            self._filters = {order_key:order_param}
        else:
            self._filters = {order_key:order_by_param}
        return queryset

    def get_ordering(self,order_param):
        self.field_sort = None
        negative_fn = lambda x:x[1:] if x.startswith('-') else '-'+x
        positive_fn = lambda x:x[1:] if x.startswith('-') else x
        desc_order = True if order_param.startswith('-') else False
        order_param = order_param[1:] if desc_order else order_param
        order_by_tuple = self.order_by_clause.get(order_param,[])
        if desc_order:
            order_by_tuple = map(negative_fn,order_by_tuple)
        if len(order_by_tuple)==1:
            self.field_sort = positive_fn(order_by_tuple[0])
        return order_by_tuple

    def get_applied_filters(self):
        return self._filters

class MutableDjangoFilterBackend(DjangoFilterBackend):
    """
    The default DjangoFilterBackend picks filter from 
    request.query_params which is immutable object. 
    So this has to be used.This filter picks additional params 
    from extra_fargs. Note the use of list object to add additional filter.
    """
    def filter_queryset(self,request,queryset,view):
        filter_class = self.get_filter_class(view, queryset)
        if isinstance(request.query_params,QueryDict):
            fargs = request.query_params.dict()
        else:
            fargs = request.query_params
        if filter_class:
            #in case of related views filter_params is set
            if view.kwargs:
                fargs.update(view.kwargs)
            filterobj=filter_class(fargs, queryset=queryset)
            qs= filterobj.qs
            self.applied_filters = filterobj.form.cleaned_data
            return qs
        return queryset

    def get_applied_filters(self):
        return self.applied_filters
